#ifndef OPAL_CODEC_ECHOSUPPRESS_H
#define OPAL_CODEC_ECHOSUPPRESS_H

#ifdef P_USE_PRAGMA
#pragma interface
#endif

#include <opal_config.h>
#include <rtp/rtp.h>

#define AGC_THRESHOLD 1000 // sample value
#define AGC_ATTACK    0.5  // percent/100, 0.0 to 1.0
#define AGC_DECAY     1.0  // percent/100, 0.0 to 1.0

// #define _ECHO_SUPPRESS_DEBUG_INFO 1

typedef enum {
	eIncomingAudio		= 0,
    eOutgoingAudio		= 1
} voipSuppressEnum;
const unsigned kSuppressTypeCount = 2;

class OpalEchoSuppress : public PObject
{
	PCLASSINFO(OpalEchoSuppress, PObject);

public:
	P_DECLARE_STREAMABLE_ENUM(	Modes,
								None,
								Fixed,
								Adaptive
	);
	typedef Modes Mode; // Backward compatibility

	struct Params
	{
		Params(
			Modes mode = Adaptive,					//  mode
			unsigned threshold		= 0,			//  threshold for fixed mode
			unsigned beginDelay		= 10,			//  10 milliseconds before silencing
			unsigned endDelay		= 400,			//  400 milliseconds before un-silencing
			unsigned adaptivePeriod = 600			//  600 millisecond window for adaptive threshold
			) :
		m_mode(mode),
			m_threshold(threshold),
			m_beginDelay(beginDelay),
			m_endDelay(endDelay),
			m_adaptivePeriod(adaptivePeriod)
		{ }

		PString AsString() const;
		void FromString(const PString & str);

		Modes    m_mode;							/// Silence detection mode
		unsigned m_threshold;						/// Threshold value if FixedEchoSuppress
		unsigned m_beginDelay;						/// milliseconds of signal needed
		unsigned m_endDelay;						/// milliseconds of silence needed
		unsigned m_adaptivePeriod;					/// millisecond window for adaptive threshold
	};

	/**@name Construction */
	//@{
	/**Create a new detector. Default clock rate is 8000.
	*/
	OpalEchoSuppress(
		const Params & newParamEcho,				///<  New parameters for silence detector
		const Params & newParamSilence				///<  New parameters for silence detector
		);
	//@}
   ~OpalEchoSuppress();

	/**@name Basic operations */
	//@{
	const PNotifier & GetOutgoingHandler() const { return outgoingHander; }
	const PNotifier & GetIncomingHandler() const { return incomingHandler; }

	bool bPushToTalk_Enabled;
	bool bAGCSpkr_Enabled;
	bool bAGCMic_Enabled;
	bool bPushToTalk_Pushed;
	void PushToTalk(bool bPushed);
	bool bHalfDuplex_Enabled;

	/**Set the silence detector parameters.
	This adjusts the silence detector "agression". The deadband and 
	adaptive periods are in ms units to work for any clock rate.
	The clock rate value is optional: 0 leaves value unchanged.
	This may be called while audio is being transferred, but if in
	adaptive mode calling this will reset the filter.
	*/
	void SetParameters(
		const Params & paramsEcho,					///< New parameters for silence detector
		const Params & paramsSilence,				///< New parameters for silence detector
		const int clockRate = 0						///< Sampling clock rate for the preprocessor
		);

	/**Get the current parameters in use.
	*/
	void GetParameters(
		Params & paramsEcho, Params & paramsSilence);

	unsigned GetEchoSuppressThreshold()			{ return levelThreshold[eIncomingAudio]; };
	unsigned GetEchoSuppressAdaptivePeriod()	{ return adaptivePeriod[eIncomingAudio] * 1000 / clockRate; };

	unsigned GetSilenceSuppressThreshold()		{ return levelThreshold[eOutgoingAudio]; };
	unsigned GetSilenceSuppressAdaptivePeriod() { return adaptivePeriod[eOutgoingAudio] * 1000 / clockRate; };

	bool IsStreamSilenced(voipSuppressEnum eStream) { return stream_silenced[eStream]; };

	/**Set the sampling clock rate for the preprocessor.
	Adusts the interpretation of time values.
	This may be called while audio is being transferred, but if in
	adaptive mode calling this will reset the filter.
	*/
	void SetClockRate(
		unsigned clockRate     ///< Sampling clock rate for the preprocessor
		);

	/**Get the current sampling clock rate.
	*/
	unsigned GetClockRate() const { return clockRate; }

	/**Get silence detection status

	The inTalkBurst value is true if packet transmission is enabled and
	false if it is being suppressed due to silence.

	The currentThreshold value is the value from 0 to 32767 which is used
	as the threshold value for 16 bit PCM data.
	*/
	Mode GetEchoStatus(
		PBoolean * isInTalkBurst,
		unsigned * currentThreshold
		) const;
	Mode GetSilenceStatus(
		PBoolean * isInTalkBurst,
		unsigned * currentThreshold
		) const;

	void ScaleStream(const BYTE * buffer, PINDEX size, float fScale);
	float OpalEchoSuppress::AGC(const BYTE * buf, PINDEX bufsize, float last_gain, short threshold, float attack, float decay);

	/**Get the average signal level in the stream.
	This is called from within the silence detection algorithm to
	calculate the average signal level of the last data frame read from
	the stream.

	The default behaviour returns UINT_MAX which indicates that the
	average signal has no meaning for the stream.
	*/
	virtual unsigned GetAverageSignalLevel(
		const BYTE * buffer,  ///<  RTP payload being detected
		PINDEX size           ///<  Size of payload buffer
		) = 0;

	/**Reset the adaptive filter
	*/
	void			AdaptiveResetEcho();
	void			AdaptiveResetSilence();

	bool			InIncomingTalkBurst() { return InTalkBurst[eIncomingAudio]; }
	bool			InOutgoingTalkBurst() { return InTalkBurst[eOutgoingAudio]; }

	unsigned		GetIncomingLevel() { return last_level[eIncomingAudio]; };
	unsigned		GetOutgoingLevel() { return last_level[eOutgoingAudio]; };

	void			SetSoftGainFactor(float gainFactor) { softGainFactor = gainFactor; };
	void			SetAGCSpkrEnabled(bool bEnabled) { bAGCSpkr_Enabled = bEnabled; };
	void			SetAGCMicEnabled(bool bEnabled) { bAGCMic_Enabled = bEnabled; }
	void			SetPushToTalkEnabled(bool bEnabled) { bPushToTalk_Enabled = bEnabled; };
	void			SetHalfDuplexEnabled(bool bEnabled) { bHalfDuplex_Enabled = bEnabled; };

	void			SendingDTMF(WORD wDuration);

protected:
	PDECLARE_NOTIFIER(RTP_DataFrame, OpalEchoSuppress, OutgoingPacket);
	PDECLARE_NOTIFIER(RTP_DataFrame, OpalEchoSuppress, IncomingPacket);

	void ProcessPacket(RTP_DataFrame & frame, bool bIncomingEcho);

	PNotifier		outgoingHander;
	PNotifier		incomingHandler;

	unsigned		clockRate;					// audio sampling rate

	Mode			mode[kSuppressTypeCount];
	unsigned		beginDelay[kSuppressTypeCount];				// #samples of signal needed
	unsigned		endDelay[kSuppressTypeCount];				// #samples of silence needed
	unsigned		adaptivePeriod[kSuppressTypeCount];			// #samples window for adaptive threshold

	unsigned		lastTimestamp[kSuppressTypeCount];			// Last timestamp received
	unsigned		receivedTime[kSuppressTypeCount];			// Signal/Silence duration received so far.
	unsigned		levelThreshold[kSuppressTypeCount];			// Threshold level for silence/signal
	unsigned		signalReceivedTime[kSuppressTypeCount];		// Duration of signal received
	unsigned		silenceReceivedTime[kSuppressTypeCount];	// Duration of silence received

	unsigned		average_sum[kSuppressTypeCount];
	unsigned		average_count[kSuppressTypeCount];

	bool			InTalkBurst[kSuppressTypeCount];
	unsigned		last_level[kSuppressTypeCount];
	bool			stream_silenced[kSuppressTypeCount];	// Indicates whether or not the stream is silenced for any reason
	float			softGainFactor;
	float           LastAGCSpkrGain;
	float           LastAGCMicGain;

	unsigned		DTMFDuration;
	unsigned		DTMFTimeStamp;

	PMutex			inUse;                 // Protects values to allow change while running

#ifdef _ECHO_SUPPRESS_DEBUG_INFO
	unsigned				level_history_level[50];
	unsigned				level_history_stream[50];
	unsigned				level_history_timestamp[50];
	DWORD					firstLoudPacketTime;
	typedef struct {
		char						szDescription[50];
		char						szMiscDescription[50];
		unsigned					uiMisc;
		char						szMiscTimeDescription[50];
		unsigned					uiMiscTime;
		voipSuppressEnum			eStream;
		unsigned					uiTickCount;
		unsigned					uiMicTime;
		unsigned					uiSpkTime;
		unsigned					uiTimeSinceLastFrame;
		unsigned					uiMicChangeTimeStamp;
		unsigned					uiReceivedTime;
	} echoDebug;
	void			AddDebugMessage( char * szDescription, char * szMiscDesc, DWORD dwMisc, char * szMiscTimeDesc, unsigned uiMiscTime, voipSuppressEnum eCurrentStream);
	void			DumpDebugMessages();
	echoDebug		debugInfo[1000];
	short			debugInfoIndex;

#endif
};


class OpalPCM16EchoSuppress : public OpalEchoSuppress
{
	PCLASSINFO(OpalPCM16EchoSuppress, OpalEchoSuppress);
public:
	/** Construct new silence detector for PCM-16/8000
	*/
	OpalPCM16EchoSuppress(
		const Params & newParamEcho,	 ///<  New parameters for silence detector
		const Params & newParamSilence ///<  New parameters for silence detector
		) : OpalEchoSuppress(newParamEcho,newParamSilence) { }

	/**@name Overrides from OpalEchoSuppress */
	//@{
	/**Get the average signal level in the stream.
	This is called from within the silence detection algorithm to
	calculate the average signal level of the last data frame read from
	the stream.

	The default behaviour returns UINT_MAX which indicates that the
	average signal has no meaning for the stream.
	*/
	virtual unsigned GetAverageSignalLevel(
		const BYTE * buffer,  ///<  RTP payload being detected
		PINDEX size           ///<  Size of payload buffer
		);
	//@}
};


extern ostream & operator<<(ostream & strm, OpalEchoSuppress::Mode mode);


#endif // OPAL_CODEC_ECHOSUPPRESS_H


/////////////////////////////////////////////////////////////////////////////
