#include <ptlib.h>
#ifdef __GNUC__
#pragma implementation "echosuppress.h"
#endif
#include <opal_config.h>
#include <codec/echosuppress.h>
#include <opal/patch.h>

#define new PNEW


// defines and constants
// #define VOIP_DEBUG_ADAPTIVE 1

const int			kAdaptiveThresholdPad					= 10;
const int			kAdaptiveThresholdEchoMax				= 90;
const int			kAdaptiveThresholdSilenceMax			= 55;
const int			kAdaptiveThresholdMin					= 15;
const int			kAdaptiveThresholdIncreaseRatio			= 8;
const int			kAdaptiveThresholdDecreaseRatio			= 3;
const int			kAdaptiveThresholdInitValue				= 50;
const int			kDTMFDurationBufferTime					= 120;
const int			kMinimumLevelForAdaptiveAdjustment		= 5;
const float			kSilencePacketScaleFactor				= 0.01;


extern "C" {
	unsigned char linear2ulaw(int pcm_val);
	int ulaw2linear(unsigned char u_val);
};

OpalEchoSuppress::~OpalEchoSuppress(){
#ifdef _ECHO_SUPPRESS_DEBUG_INFO
	DumpDebugMessages();
#endif
}

#ifdef _ECHO_SUPPRESS_DEBUG_INFO
void OpalEchoSuppress::DumpDebugMessages()
{
	unsigned prevTick = debugInfo[0].uiTickCount;
	for ( int i=0; i<debugInfoIndex; i++ ) {
		PTRACE(1,
																debugInfo[i].szDescription									<<
					"  Stream="									<<	debugInfo[i].eStream									<<
					"  TickTm ="								<< (debugInfo[i].uiTickCount - prevTick )					<<
					"  LastMicTm="								<< (debugInfo[i].uiMicTime  * 1000 / clockRate )			<<
					"  LastSpkTm="								<< (debugInfo[i].uiSpkTime * 1000 / clockRate )				<<
					// "  TmSinceLast="							<< (debugInfo[i].uiTimeSinceLastFrame * 1000 / clockRate )	<<
					// "  MicChangeTm="							<< (debugInfo[i].uiMicChangeTimeStamp * 1000 / clockRate )	<<
					"  ReceiveTm="								<< (debugInfo[i].uiReceivedTime * 1000 / clockRate )		<<
					" " << debugInfo[i].szMiscTimeDescription	<< (debugInfo[i].uiMiscTime * 1000 / clockRate )			<<
					" " << debugInfo[i].szMiscDescription		<<	debugInfo[i].uiMisc );
		prevTick = debugInfo[i].uiTickCount;
	}
}

void OpalEchoSuppress::AddDebugMessage
(
	char *					szDescription,
	char *					szMiscDesc,
	DWORD					dwMisc,
	char *					szMiscTimeDesc,
	unsigned				uiMiscTime,
	voipSuppressEnum		eCurrentStream
)
{
	strcpy( debugInfo[debugInfoIndex].szDescription, szDescription );
	strcpy( debugInfo[debugInfoIndex].szMiscDescription, szMiscDesc );
	strcpy( debugInfo[debugInfoIndex].szMiscTimeDescription, szMiscTimeDesc );
	debugInfo[debugInfoIndex].uiTickCount				= GetTickCount();
	debugInfo[debugInfoIndex].uiMicTime					= lastTimestamp[eOutgoingAudio];
	debugInfo[debugInfoIndex].uiSpkTime					= lastTimestamp[eIncomingAudio];
	debugInfo[debugInfoIndex].eStream					= eCurrentStream;
	debugInfo[debugInfoIndex].uiTimeSinceLastFrame		= 0;
	debugInfo[debugInfoIndex].uiMicChangeTimeStamp		= 0; // micChangeTimeStamp;
	debugInfo[debugInfoIndex].uiReceivedTime			= receivedTime[eCurrentStream];
	debugInfo[debugInfoIndex].uiMiscTime				= uiMiscTime;
	debugInfo[debugInfoIndex].uiMisc					= dwMisc;
	debugInfoIndex++;
	if ( debugInfoIndex > 999 ) debugInfoIndex = 0;
}
#endif

OpalEchoSuppress::OpalEchoSuppress(const Params & theParamEcho, const Params & theParamSilence)
	:
	outgoingHander ( PCREATE_NOTIFIER( OutgoingPacket )),
	incomingHandler( PCREATE_NOTIFIER( IncomingPacket )),
	clockRate(8000),
	bPushToTalk_Pushed(false),
	bPushToTalk_Enabled(false),
	bAGCSpkr_Enabled(true),
	bAGCMic_Enabled(false),
	softGainFactor(100),
	DTMFDuration(0),
	DTMFTimeStamp(0),
	bHalfDuplex_Enabled(false)
{
	LastAGCSpkrGain = 1.0;
	LastAGCMicGain = 1.0;

	stream_silenced[0] = false;
	stream_silenced[1] = false;

#ifdef _ECHO_SUPPRESS_DEBUG_INFO
	firstLoudPacketTime = 0;
	debugInfoIndex = 0;
#endif

	// Initialise the adaptive threshold variables.
	SetParameters(theParamEcho,theParamSilence);

	PTRACE(4, "EchoSuppress\tHandler created");
}

void OpalEchoSuppress::AdaptiveResetEcho()
{
	if (mode[eIncomingAudio] != Adaptive)
		return;

	// Initialise threshold level
	levelThreshold[eIncomingAudio]			= kAdaptiveThresholdInitValue;

	// Initialise the adaptive threshold variables.
	signalReceivedTime[eIncomingAudio]		= 0;
	silenceReceivedTime[eIncomingAudio]		= 0;
	average_sum[eIncomingAudio]				= 0;
	average_count[eIncomingAudio]			= 0;

	// Restart in silent mode
	InTalkBurst[eIncomingAudio]				= false;

	lastTimestamp[eIncomingAudio]			= 0;
	receivedTime[eIncomingAudio]			= 0;
}

void OpalEchoSuppress::AdaptiveResetSilence()
{
	if (mode[eOutgoingAudio] != Adaptive)
		return;

	// Initialise threshold level
	levelThreshold[eOutgoingAudio]		= kAdaptiveThresholdInitValue;

	// Initialise the adaptive threshold variables.
	signalReceivedTime[eOutgoingAudio]	= 0;
	silenceReceivedTime[eOutgoingAudio]	= 0;
	average_sum[eOutgoingAudio]			= 0;
	average_count[eOutgoingAudio]		= 0;

	// Restart in silent mode
	InTalkBurst[eOutgoingAudio]			= false;
	lastTimestamp[eOutgoingAudio]		= 0;
	receivedTime[eOutgoingAudio]		= 0;
}

void OpalEchoSuppress::PushToTalk(bool bPushed)
{
	PTRACE(2, "Echo Suppress\tPush to talk - " << bPushed );
	bPushToTalk_Pushed = bPushed;
}

void OpalEchoSuppress::SetParameters(const Params & newParamEcho, const Params & newParamSilence, const int rate /*= 0*/)
{
	PWaitAndSignal mutex(inUse);
	
	if (rate)
		clockRate = rate;
	
	mode[eIncomingAudio]				= newParamEcho.m_mode;

	beginDelay[eIncomingAudio]			= newParamEcho.m_beginDelay			* clockRate / 1000;
	endDelay[eIncomingAudio]			= newParamEcho.m_endDelay			* clockRate / 1000;
	adaptivePeriod[eIncomingAudio]		= newParamEcho.m_adaptivePeriod		* clockRate / 1000;

	// fixed mode - use the passed threshold
	if (mode[eIncomingAudio] == Fixed)
		levelThreshold[eIncomingAudio]	= newParamEcho.m_threshold;		// note: this value compared to uLaw encoded signal level

	// else adaptive
	else
		AdaptiveResetEcho();

	mode[eOutgoingAudio]				= newParamSilence.m_mode;
	beginDelay[eOutgoingAudio]			= newParamSilence.m_beginDelay		* clockRate / 1000;
	endDelay[eOutgoingAudio]			= newParamSilence.m_endDelay		* clockRate / 1000;
	adaptivePeriod[eOutgoingAudio]		= newParamSilence.m_adaptivePeriod	* clockRate / 1000;

	// fixed mode - use the passed threshold
	if (mode[eOutgoingAudio] == Fixed)
		levelThreshold[eOutgoingAudio]	= newParamSilence.m_threshold;		// note: this value compared to uLaw encoded signal level

	// else adaptive
	else
		AdaptiveResetSilence();

	PTRACE(2,	"Echo\tParameters set: "
				"mode="			<< mode[eIncomingAudio]				<< ", "
				"threshold="	<< levelThreshold[eIncomingAudio]	<< ", "
				"silencedb="	<< endDelay[eIncomingAudio]			<< " samples, "
				"signaldb="		<< beginDelay[eIncomingAudio]		<< " samples, "
				"period="		<< adaptivePeriod					<< " samples, "
				"clockrate="	<< clockRate);

	PTRACE(2,	"Silence\tParameters set: "
				"mode="			<< mode[eOutgoingAudio]				<< ", "
				"threshold="	<< levelThreshold[eOutgoingAudio]	<< ", "
				"silencedb="	<< endDelay[eOutgoingAudio]			<< " samples, "
				"signaldb="		<< beginDelay[eOutgoingAudio]		<< " samples, "
				"period="		<< adaptivePeriod[eOutgoingAudio]	<< " samples, "
				"clockrate="	<< clockRate);
}

void OpalEchoSuppress::SetClockRate(unsigned rate)
{
	PWaitAndSignal mutex(inUse);

	beginDelay[eIncomingAudio]			= beginDelay[eIncomingAudio]		* 1000 / clockRate * rate / 1000;
	endDelay[eIncomingAudio]			= endDelay[eIncomingAudio]			* 1000 / clockRate * rate / 1000;
	adaptivePeriod[eIncomingAudio]		= adaptivePeriod[eIncomingAudio]	* 1000 / clockRate * rate / 1000;

	beginDelay[eOutgoingAudio]			= beginDelay[eOutgoingAudio]		* 1000 / clockRate * rate / 1000;
	endDelay[eOutgoingAudio]			= endDelay[eOutgoingAudio]			* 1000 / clockRate * rate / 1000;
	adaptivePeriod[eOutgoingAudio]		= adaptivePeriod[eOutgoingAudio]	* 1000 / clockRate * rate / 1000;

	clockRate = rate;

	if (mode[eIncomingAudio] == Adaptive) 
		AdaptiveResetEcho();

	if (mode[eOutgoingAudio] == Adaptive)
		AdaptiveResetSilence();
}

void OpalEchoSuppress::GetParameters(Params & paramsEcho, Params & paramsSilence)
{
	// echo params
	paramsEcho.m_mode				= mode[eIncomingAudio];
	paramsEcho.m_threshold			= levelThreshold[eIncomingAudio]	* 1000 / clockRate;
	paramsEcho.m_beginDelay			= beginDelay[eIncomingAudio]		* 1000 / clockRate;
	paramsEcho.m_endDelay			= endDelay[eIncomingAudio]			* 1000 / clockRate;
	paramsEcho.m_adaptivePeriod		= adaptivePeriod[eIncomingAudio]	* 1000 / clockRate;

	// silence detect params
	paramsSilence.m_mode			= mode[eOutgoingAudio];
	paramsSilence.m_threshold		= levelThreshold[eOutgoingAudio]	* 1000 / clockRate;
	paramsSilence.m_beginDelay		= beginDelay[eOutgoingAudio]		* 1000 / clockRate;
	paramsSilence.m_endDelay		= endDelay[eOutgoingAudio]			* 1000 / clockRate;
	paramsSilence.m_adaptivePeriod	= adaptivePeriod[eOutgoingAudio]	* 1000 / clockRate;
}

PString OpalEchoSuppress::Params::AsString() const
{
	return PSTRSTRM(m_mode << ',' << m_threshold << ',' << m_beginDelay << ',' << m_endDelay << ',' << m_adaptivePeriod);
}

void OpalEchoSuppress::Params::FromString(const PString & str)
{
	PStringArray params = str.Tokenise(',');
	switch (params.GetSize()) {
		default :
		case 5 :
			m_adaptivePeriod = params[4].AsUnsigned();
		case 4 :
			m_endDelay = params[3].AsUnsigned();
		case 3 :
			m_beginDelay = params[2].AsUnsigned();
		case 2 :
			m_threshold = params[1].AsUnsigned();
		case 1 :
			m_mode = params[0].As<OpalEchoSuppress::Modes>(m_mode);
		case 0 :
			break;
	}
}

OpalEchoSuppress::Mode OpalEchoSuppress::GetEchoStatus(PBoolean * isInReceiveTalkBurst, unsigned * currentThreshold) const
{
	// return whether or not we're currently recieving audio above our threshold
	if (isInReceiveTalkBurst != NULL) {
		*isInReceiveTalkBurst = InTalkBurst[eIncomingAudio];
	}

	if (currentThreshold != NULL)
		*currentThreshold = ulaw2linear((BYTE)(levelThreshold[eIncomingAudio] ^ 0xff));

	return mode[eIncomingAudio];
}

OpalEchoSuppress::Mode OpalEchoSuppress::GetSilenceStatus(PBoolean * isInTalkBurst, unsigned * currentThreshold) const
{
	// return whether or not we're currently recieving audio above our threshold
	if (isInTalkBurst != NULL) {
		*isInTalkBurst = InTalkBurst[eOutgoingAudio];
	}

	if (currentThreshold != NULL)
		*currentThreshold = ulaw2linear((BYTE)(levelThreshold[eIncomingAudio] ^ 0xff));

	return mode[eIncomingAudio];
}

#pragma warning ( disable : 4244 ) // conversion from float to int
void OpalEchoSuppress::ScaleStream(const BYTE * buffer, PINDEX size, float fScale)
{
	register int resultsample;

	// no-op
	if ( fScale == 1.0f )
		return;

	PINDEX samples = size/2;
	short * pcm = (short *)buffer;
	const short * end = pcm + samples;

    /* scale the input signal to achieve the required output average */
	while (pcm != end) {
		resultsample = *pcm * fScale;

		// Watch out for overflow. Here, overflow is converted 
		// to slightly less appalling sound; clipping.
		if (resultsample > 32767)
			*pcm = 32767;
		else if (resultsample < -32767)
			*pcm = -32767;
		else
			*pcm = resultsample;			

		pcm++;
	}
}

/**************************************************************************/
/* This function performs automatic gain control on the data stream.      */
/*                                                                        */
/* This function assumes each audio sample is 16 bits.                    */
/*                                                                        */
/* Function arguments:                                                    */
/*                                                                        */
/* buf- audio data buffer                                                 */
/* bufsize- number of bytes in the audio data buffer                      */
/* last_gain- gain setting of previous audio data buffer of the stream    */
/* threshold- minimum audio sample value to perform gain control          */
/* attack- how fast the gain is allowed increase; 0.0 to 1.0              */
/* decay- how fast the gain is allowed to fall; 0.0 to 1.0                */
/*                                                                        */
/* Comments:                                                              */
/*                                                                        */
/* The threshold sets at what point the AGC will act. Audio data below    */
/* the threshold will not be altered.                                     */
/*                                                                        */
/* The attack is used to control how fast the gain can rise. If the gain  */
/* is allowed to rise too fast, the normal variations of amplitude may be */
/* lost. If the attack is set too low, the audio will be consistantly     */
/* quiet.                                                                 */
/*                                                                        */
/* The decay is used to control how fast the gain can lower. If the decay */
/* drops too slowly, normal fluctionations in amplitude may begin to      */
/* clip. If the decay is set too fast, less overall gain will be          */
/* available.                                                             */
/*                                                                        */
/* The last_gain argument is the value returned from this function for    */
/* the previous audio buffer. This is used with atack, and decay to       */
/* reduce distortion by smoothing the gain transitions from audio buffer  */
/* to audio buffer.                                                       */
/**************************************************************************/
float OpalEchoSuppress::AGC(const BYTE * buf, PINDEX bufsize, float last_gain, short threshold, float attack, float decay)
{
	register short* sample;
	short max;
	float gain;
	
	sample = (short*)buf;
	max = 0;
	
	// Find the max
	for (sample = (short*)buf; sample < (short*)(buf + bufsize); sample++) {
		if (*sample > 0)
			if (*sample > max) max = *sample;
		else
			if (*sample < -max) max = -(*sample);
	}
	
	// Ignore very low volumes
	if (max < threshold)
		gain = 1.0;
	else
		gain = 32766.0 / max; // Give us a margin of 1 to avoid clipping
		
	// Add some hysteresis in rising gain, and falling gain
	if (gain > last_gain)
		gain = ((gain - last_gain) * attack) + last_gain;
	else
		gain = ((gain - last_gain) * decay) + last_gain;
	
	// Apply the corrected gain
	ScaleStream(buf, bufsize, gain);

	return gain;
}

void OpalEchoSuppress::ProcessPacket(RTP_DataFrame & frame, bool bIncomingStream)
{
	PWaitAndSignal mutex(inUse);

	int iIndex;
	unsigned int maxThreshold;
	bool bTalkBurstSet = false;
	
	// if incoming audio
	if ( bIncomingStream ) {
		iIndex = eIncomingAudio;
		maxThreshold = kAdaptiveThresholdEchoMax;
	}

	// else outgoing audio
	else {
		iIndex = eOutgoingAudio;
		maxThreshold = kAdaptiveThresholdSilenceMax;
	}

	// average is absolute value up to 32767
	unsigned level = 0;

	// if we have a payload
	if ( frame.GetPayloadSize() != 0 ) {

		// incoming stream - AGC
		if ( bIncomingStream ) {
#if SCALE_CALL_VOLUME
			ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), softGainFactor);
#endif
			if (bAGCSpkr_Enabled)
				LastAGCSpkrGain = AGC(frame.GetPayloadPtr(), frame.GetPayloadSize(), LastAGCSpkrGain, AGC_THRESHOLD, AGC_ATTACK, AGC_DECAY);
		}

		// outgoing stream - AGC
		else {
			if (bAGCMic_Enabled)
				LastAGCMicGain = AGC(frame.GetPayloadPtr(), frame.GetPayloadSize(), LastAGCMicGain, AGC_THRESHOLD, AGC_ATTACK, AGC_DECAY);
		}

		// calculate the payload level
		level = GetAverageSignalLevel(frame.GetPayloadPtr(), frame.GetPayloadSize());

		// can never have average signal level that high, this indicates that the
		// hardware cannot do echo suppression.
		if (level == UINT_MAX) {
			last_level[iIndex] = 0;
			stream_silenced[iIndex] = false;
			return;
		}

		// convert to a logarithmic scale - use uLaw which is complemented
		level = linear2ulaw(level) ^ 0xff;
	}

	// store the level
	last_level[iIndex] = level;
	stream_silenced[iIndex] = false; // default to not, over-ride if we silence the packet below

#ifdef _ECHO_SUPPRESS_DEBUG_INFO
	memcpy(&level_history_level[0],&level_history_level[1],sizeof(unsigned)*49);
	memcpy(&level_history_stream[0],&level_history_stream[1],sizeof(unsigned)*49);
	memcpy(&level_history_timestamp[0],&level_history_timestamp[1],sizeof(unsigned)*49);
	level_history_level[49]			= level;
	level_history_stream[49]		= iIndex;
	level_history_timestamp[49]		= frame.GetTimestamp();

	/* this code is used to figure out how long it takes a sounds burst to get from the speaker input packet, out the speaker, into the mic, and into
	   this code again for the microphone packet. */
	if ( level > 80 ) {
		if ( firstLoudPacketTime == 0 ) {
			if ( iIndex == eIncomingAudio ) {
				firstLoudPacketTime = GetTickCount();
				// PTRACE(1, "First loud Packet!!! stream = " << (( eCurrentStream == eSpeakerStream ) ? "Speaker" : "Microphone" ));
			}
		} else if (( iIndex == eOutgoingAudio ) && ( firstLoudPacketTime != 1 )) {
			DWORD secondLoudPacketTime = GetTickCount();
			DWORD dwTimeDiff = secondLoudPacketTime - firstLoudPacketTime;

			AddDebugMessage( "Spk-2-mic:", "delay:", dwTimeDiff, "", 0, (voipSuppressEnum) iIndex);

			firstLoudPacketTime = 1;
			// findings:
			// there's about 100ms delay from the time the speaker plays a sound to the time the mic picks it up.  That directly relates to the buffer
			// set in RTIVoIPManager.cpp with the call to SetSoundChannelBufferTime(), currently set to 60ms.  If that is increased to 360ms then there's
			// about a 200-400ms delay, not sure why that's not consistent.
		}
	} 
#endif

	// decide whether or not we're in a talk burst, and clear the payload when appropriate...


	// if incoming stream
	if ( bIncomingStream ) {

		// if push-to-talk is enabled and the talk button isn't pressed, we're in a talk burst
		if ( bPushToTalk_Enabled && !bPushToTalk_Pushed ) {

			// we're in a talk burst
			InTalkBurst[eIncomingAudio] = true;

			// set a flag indicating that we figured out talk burst already
			bTalkBurstSet = true;
		}

		// if the talk button is pressed incoming audio is silenced (even if PTT is disabled)
		if ( bPushToTalk_Pushed ) {

			// we are not in an incoming talk burst
			InTalkBurst[eIncomingAudio] = false;

			// set a flag indicating that we figured out talk burst already
			bTalkBurstSet = true;

			// silence the packet
			ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);

			stream_silenced[iIndex] = true;

			// we're going to return - in-talk-burst is false and we've cleared the incoming audio packet, nothing
			// further to do.  Right?  what about adaptive?
			return;
		}
	}

	// else outgoing stream
	else {

		// if talk button is pressed we're in a talk burst regardless
		if ( bPushToTalk_Pushed ) {
			
			// we're in a talk burst
			InTalkBurst[eOutgoingAudio] = true;

			// set a flag indicating that we figured out talk burst already
			bTalkBurstSet = true;
		}

		// else the talk button isn't being pressed
		else {

			// if push-to-talk is enabled, and the Talk button isn't being pushed, we're not in a talk burst
			if ( bPushToTalk_Enabled ) {

				// we're not in a talk burst
				InTalkBurst[eOutgoingAudio] = false;

				// set a flag indicating that we figured out talk burst already
				bTalkBurstSet = true;
			}

			// else if echo suppression is on and the other guy is talking
			else if ( ( mode[eIncomingAudio] != None) && ( InTalkBurst[eIncomingAudio] )) {

				// we're not in a talk burst
				InTalkBurst[eOutgoingAudio] = false;

				// set a flag indicating that we figured out talk burst already
				bTalkBurstSet = true;
			}
		}

		// if we've set the talk burst flag already, and we're not in a talk burst, silence the packet
		if ( bTalkBurstSet && !InTalkBurst[eOutgoingAudio] ) {

			// reset the last time stamp since we're interrupting the silence suppression
			// mike todo lastTimestamp[iIndex] = 0;

			// silence the packet
			ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);

			stream_silenced[iIndex] = true;

			// we're going to return - no need to continue with silence suppression code since we've already said that
			// we're not in a talk burst and have silenced the frame, so we don't want it over-riding that
			return;
		}
	}

	
	// get the current time
	unsigned thisTimestamp = frame.GetTimestamp();


	// if we're processing an in-line DTMF code we want to suppress incoming audio and ensure that 
	// we don't suppress outgoing audio
	if ( DTMFDuration > 0 ) {

		// if this is the first packet that we're sending DTMF for, set the timestamp
		if ( DTMFTimeStamp == 0 ) {
			DTMFTimeStamp = thisTimestamp;
		}
		else {
			// calculate time since the DTMF start
			unsigned timeSinceDTMFStart;
			if ( thisTimestamp > DTMFTimeStamp )
				timeSinceDTMFStart = thisTimestamp - DTMFTimeStamp;
			else
				timeSinceDTMFStart = 0;

			// if the DTMF char is already sent, we're done, clear flags
			if ( timeSinceDTMFStart > DTMFDuration ) {
				DTMFDuration = 0;
				DTMFTimeStamp = 0;
				timeSinceDTMFStart = 0;
			}

			// else we're still sending the DTMF char
			else {
			
				// if incoming stream, silence it while we're sending the code
				if ( bIncomingStream ) {
				
					// we are not in an incoming talk burst
					InTalkBurst[eIncomingAudio] = false;

					// silence the packet
					ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);

					stream_silenced[eIncomingAudio] = true;
				}

				// else outgoing stream, return so that we don't silence the packet below
				else {
					// send the audio
					InTalkBurst[eOutgoingAudio] = true;
					return;
				}
			}
		}
	}


	// if push-to-talk is enabled we'll be done now
	if ( bPushToTalk_Enabled ) {
		return;
	}


	// if no suppression, our job is done here
	if (mode[iIndex] == None) {
		return;
	}


	// if this is the mic stream and half-duplex is enabled we're done since we only get silenced based on the speaker
	// stream above
	if ( !bIncomingStream && bHalfDuplex_Enabled ) {
		return;
	}



	// for outgoing audio, the only way we get here is if silence suppression is on (fixed or adaptive), push-to-talk is disabled,
	// and the talk button is not being pressed

	// for incoming audio, the only way we get here is if echo suppression is on (fixed or adaptive), push-to-talk is disabled,
	// and the talk butto 


	// first packet, keep the current time stamp and send the packet
	if (lastTimestamp[iIndex] == 0) {
		
		ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);

		// clear last level value
		last_level[iIndex] = 0;	
		lastTimestamp[iIndex] = thisTimestamp;
		stream_silenced[iIndex] = false;

		return;
	}


	// calculate the time since the last packet
	unsigned timeSinceLastFrame = thisTimestamp - lastTimestamp[iIndex];
	lastTimestamp[iIndex] = thisTimestamp;


	// if we haven't set the talk burst flag above, we'll do so now based on current level
	if ( !bTalkBurstSet ) {

		// if the current level is above the threshold, we're in a talk burst
		PBoolean currentTalkBurst = level > levelThreshold[iIndex];

		// if no change, still talking or still silent, reset frame timer
		if (InTalkBurst[iIndex] == currentTalkBurst)
			receivedTime[iIndex] = 0;

		// else there's a change in state
		else {

#ifdef _ECHO_SUPPRESS_DEBUG_INFO
			// if ( receivedTime[iIndex] <= 0 ) {
			if ( bIncomingStream ) {
				AddDebugMessage( "Echo Chng 1:", "New Burst:", currentTalkBurst, "", 0, (voipSuppressEnum) iIndex);
			} else {
				AddDebugMessage( "Sil Chng  1:", "New Burst:", currentTalkBurst, "", 0, (voipSuppressEnum) iIndex);
			}
			// }
#endif
			// increment the time since the first packet of the same level
			receivedTime[iIndex] += timeSinceLastFrame;

			// if have had enough consecutive frames talking/silent, swap modes
			if (receivedTime[iIndex] >= (InTalkBurst[iIndex] ? (unsigned)endDelay[iIndex] : (unsigned)beginDelay[iIndex]))
			{
				// swap modes
				InTalkBurst[iIndex] = !InTalkBurst[iIndex];

				PTRACE(2, (bIncomingStream ? "Echo    " : "Silence ") << "Suppress\tTransition: " << (InTalkBurst[iIndex] ? "Talk" : "Silent") << " level=" << level << " threshold=" << levelThreshold[eOutgoingAudio]);
#ifdef _ECHO_SUPPRESS_DEBUG_INFO
				if ( bIncomingStream ) {
					AddDebugMessage( "Echo Chng 2:", "New Burst:", InTalkBurst[iIndex], "", 0, (voipSuppressEnum) iIndex);
				} else {
					AddDebugMessage( "Sil Chng  2:", "New Burst:", InTalkBurst[iIndex], "", 0, (voipSuppressEnum) iIndex);
				}
#endif

				// if we had talk/silence transition restart adaptive threshold measurements
				signalReceivedTime[eOutgoingAudio] = 0;
				silenceReceivedTime[eOutgoingAudio] = 0;
				average_sum[eOutgoingAudio] = 0;
				average_count[eOutgoingAudio] = 0;

				// if we just have moved to sending a talk burst, set the RTP marker
				if (InTalkBurst[iIndex])
					frame.SetMarker(true);
			}
		}
	}

	
	// now our talk-burst flag is updated, figure out what to do with the packet...



	// if this is the speaker stream and half-duplex is enabled and we're not talking, silence the packet
	if ( bIncomingStream && bHalfDuplex_Enabled && !InTalkBurst[iIndex] ) {
		ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);
		stream_silenced[iIndex] = true;
	}

	
	// if outgoing stream
	if ( !bIncomingStream ) {

		// if we're currently quiet, silence the frame
		if ( !InTalkBurst[eOutgoingAudio] ) {
			ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);
			stream_silenced[iIndex] = false; // we won't set it to true here since it's not really suppressed
		}
	}


	// if fixed mode we're done here, the rest of the code is related to adaptive mode
	if (mode[iIndex] == Fixed) {
		return;
	}


	// adaptive suppression...

	if (levelThreshold[iIndex] == 0) {
		if (level > 1) {
			// bootstrap condition, use first frame level as level
			levelThreshold[iIndex] = level / 2;
			PTRACE(2, (bIncomingStream ? "Echo    " : "Silence ") << "Suppress\tThreshold initialised to: " << levelThreshold[iIndex]);
		}

		// if outgoing stream, clear the audio
		if ( !bIncomingStream ) {
			ScaleStream(frame.GetPayloadPtr(), frame.GetPayloadSize(), kSilencePacketScaleFactor);
			stream_silenced[iIndex] = true;
		}
		return;
	}

	// count the number of silent and signal frames and calculate min/max
	
	// in a talk burst
	if (InTalkBurst[iIndex]) {
		signalReceivedTime[iIndex] = signalReceivedTime[iIndex] + timeSinceLastFrame;
	}

	// else silent
	else {
		silenceReceivedTime[iIndex] = silenceReceivedTime[iIndex] + timeSinceLastFrame;
	}

	if ( level > kMinimumLevelForAdaptiveAdjustment ) {
		if ( average_sum[iIndex] > ( UINT_MAX - 1000 )) {
			average_sum[iIndex] = 0;
			average_count[iIndex] = 0;
		}
		average_sum[iIndex] += level;
		average_count[iIndex]++;
	}

	// see if we have had enough frames to look at proportions of silence/signal
	if ((signalReceivedTime[iIndex] + silenceReceivedTime[iIndex]) > adaptivePeriod[iIndex]) {

		/* Now we have had a period of time to look at some average values we can
		make some adjustments to the threshold. There are three cases:
		*/

		// calculate the average level for this period
		unsigned average_level = 15;
		if ( average_count[iIndex] > 0 )
			average_level = (average_sum[iIndex] / average_count[iIndex]) + kAdaptiveThresholdPad;

		// calculate the delta
		int delta = (((int)average_level) - ((int)levelThreshold[iIndex]));

		// noise frame - increase threshold - a bit slower
		if (signalReceivedTime[iIndex] >= adaptivePeriod[iIndex]) {
			delta = delta / kAdaptiveThresholdIncreaseRatio;
			if ( delta < 1 ) 
				delta = 1;
		}

		// quiet frame - decrease threshold - a bit faster
		else if (silenceReceivedTime[iIndex] >= adaptivePeriod[iIndex]) {
			delta = delta / kAdaptiveThresholdDecreaseRatio;
		}

		// hovering around threshold - creep up the threshold
		else if (signalReceivedTime[iIndex] > silenceReceivedTime[iIndex]) {
			delta = 1;
		}

		// adjust the threshold by the calculated delta
		levelThreshold[iIndex] += delta;

		PTRACE(2, (bIncomingStream ? "Echo    " : "Silence ") 
					<< "Suppress\tThreshold changed to: " << levelThreshold[iIndex]
					<< " delta=" << delta 
					<< " avg=" << average_level
					<< " sigTm=" << signalReceivedTime[iIndex]
					<< " silTm=" << silenceReceivedTime[iIndex]);

		// we'll cap the threshold - since at some points it's obvious that the level is noisy
		if ( levelThreshold[iIndex] > maxThreshold )
			levelThreshold[iIndex] = maxThreshold;

		// we'll also force a min threshold.  For incoming audio, the other party might be sending nothing
		// if silence suppressed, so we don't want our threshold getting too low.  For silence suppression
		// there's a level that's obviously silent
		if ( levelThreshold[iIndex] < kAdaptiveThresholdMin )
			levelThreshold[iIndex] = kAdaptiveThresholdMin;

		// reset values
		signalReceivedTime[iIndex] = 0;
		silenceReceivedTime[iIndex] = 0;
		average_sum[iIndex] = 0;
		average_count[iIndex] = 0;
	}
}

// We're sending an audio packet to the other party.  We'll empty the packet (send silence) in the following cases:
//
//  - The other party is talking:
//			- if the incoming audio level is above a threshold
//			- handled in IncomingPacket()
//  - This party isn't talking:
//			- if the outgoing audio level is below a threshold
//			- handled here, in OutgoingPacket()
//
void OpalEchoSuppress::OutgoingPacket(RTP_DataFrame& frame, P_INT_PTR)
{
	ProcessPacket(frame, false);
}

// We're receiving an audio packet from the other party.  We'll analyze it to see if the audio level
// is above a threshold.  If it is we'll set a flag indicating that the other party is talking in which
// case we don't want to send audio because it'll cause echo, picking up the audio from the speaker
// note that we're also handling push to talk separately
void OpalEchoSuppress::IncomingPacket(RTP_DataFrame & frame, P_INT_PTR)
{
	ProcessPacket(frame, true);
}

unsigned OpalPCM16EchoSuppress::GetAverageSignalLevel(const BYTE * buffer, PINDEX size)
{
	// Calculate the average signal level of this frame
	int sum = 0;
	PINDEX samples = size/2;
	const short * pcm = (const short *)buffer;
	const short * end = pcm + samples;
	while (pcm != end) {
		if (*pcm < 0)
			sum -= *pcm++;
		else
			sum += *pcm++;
	}

	return sum/samples;
}


// called to indicate that a DTMF character is going to be sent for the given duration, so
// silence incoming packets and don't silence outgoing packets for that duration
void OpalEchoSuppress::SendingDTMF(WORD wDuration)
{
	DTMFDuration = wDuration + kDTMFDurationBufferTime;
}
